/**************************************************************************/
/* Title:       dagblks                                                   */
/*------------------------------------------------------------------------*/
/* Version: Amended:  By:            Changes:                             */
/*------------------------------------------------------------------------*/
/* 0.1      14/9/92   Bill Oldfield   started DAGBSMC                     */
/* 0.2      16/9/92   Bill Oldfield   added DAGDECC block (DAG decode)    */
/* 0.3      17/9/92   Bill Oldfield   added DAGINCC (incrementer)         */
/* 0.4      21/9/92   Bill            added vidlast and removed in2c buff */
/* 0.5      22/9/92   Bill            added test mode                     */
/* 0.6      23/9/92   Bill            split up decode parts               */
/* 0.7      25/9/92   Bill            inverted anb's to bna's             */
/* 0.8      29/9/92   Bill            added Noe_cco and Noe_vid           */
/* 0.9      29/9/92   Bill            added SIGGENC block                 */
/* 1.0      5/10/92   Bill            changed flybks to flybi             */
/* 1.1     12/10/92   Bill            added new mode to SIGGENC           */
/* 1.2     24/11/92   Bill            added LO(dmagoing) into sel_w_*     */
/* 1.3     21/12/92   Bill            changed Noe_cin,Noe_vin,Noe_vst,    */
/*                                    sb_cc,sb_vc,sd_vc,Ndr_w2c           */
/* 1.4     17/02/93   Bill - changes made for IOMDplus                    */
/* 1.5      5/02/93   Bill - added dma_go_vis so that buses driven during */
/*                           refresh                                      */
/* 1.6     19/03/93   Bill - added dma_go_vic for endbus drive during     */
/*                           cursor                                       */
/**************************************************************************/

#include <Handy.h>

#define UNDEF_HIGH 3

#define IO   1
#define SND0 5
#define SND1 3
#define CUR  6
#define VID  4

#define sel_w_reg(n)   ((la_val==(n)) && HI(dagreg) && HI(lwrite) && LO(dmagoing))
#define sel_r_reg(n)   ((la_val==(n)) && HI(dagreg) && LO(lwrite))

#define write_reg(n)   ( HI(prog) && sel_w_reg(n) )
#define  read_reg(n)   ( HI(prog) && sel_r_reg(n) )

#define dma_go_i0a     (HI(dmagoing) && (dmach_val==IO) && (ioch_val==0) && LO(bna_i0))
#define dma_go_i0b     (HI(dmagoing) && (dmach_val==IO) && (ioch_val==0) && HI(bna_i0))
#define dma_go_i1a     (HI(dmagoing) && (dmach_val==IO) && (ioch_val==1) && LO(bna_i1))
#define dma_go_i1b     (HI(dmagoing) && (dmach_val==IO) && (ioch_val==1) && HI(bna_i1))
#define dma_go_i2a     (HI(dmagoing) && (dmach_val==IO) && (ioch_val==2) && LO(bna_i2))
#define dma_go_i2b     (HI(dmagoing) && (dmach_val==IO) && (ioch_val==2) && HI(bna_i2))
#define dma_go_i3a     (HI(dmagoing) && (dmach_val==IO) && (ioch_val==3) && LO(bna_i3))
#define dma_go_i3b     (HI(dmagoing) && (dmach_val==IO) && (ioch_val==3) && HI(bna_i3))
#define dma_go_s0a     (HI(dmagoing) && (dmach_val==SND0)                && LO(bna_s0))
#define dma_go_s0b     (HI(dmagoing) && (dmach_val==SND0)                && HI(bna_s0))
#define dma_go_s1a     (HI(dmagoing) && (dmach_val==SND1)                && LO(bna_s1))
#define dma_go_s1b     (HI(dmagoing) && (dmach_val==SND1)                && HI(bna_s1))
#define dma_go_cur     (HI(dmagoing) && (dmach_val==CUR)                              )
#define dma_go_vid     (HI(dmagoing) && (dmach_val==VID)                              )
#define dma_go_vis     (HI(dmagoing) && ((dmach_val==VID) || (dmach_val==2) || (dmach_val==0)))
#define dma_go_vic     (HI(dmagoing) && ((dmach_val==VID) || (dmach_val==2) || (dmach_val==0) || (dmach_val==6)))
#define end_dma_i0a    (dma_go_i0a && HI(ramdmaend))
#define end_dma_i0b    (dma_go_i0b && HI(ramdmaend))
#define end_dma_i1a    (dma_go_i1a && HI(ramdmaend))
#define end_dma_i1b    (dma_go_i1b && HI(ramdmaend))
#define end_dma_i2a    (dma_go_i2a && HI(ramdmaend))
#define end_dma_i2b    (dma_go_i2b && HI(ramdmaend))
#define end_dma_i3a    (dma_go_i3a && HI(ramdmaend))
#define end_dma_i3b    (dma_go_i3b && HI(ramdmaend))
#define end_dma_s0a    (dma_go_s0a && HI(ramdmaend))
#define end_dma_s0b    (dma_go_s0b && HI(ramdmaend))
#define end_dma_s1a    (dma_go_s1a && HI(ramdmaend))
#define end_dma_s1b    (dma_go_s1b && HI(ramdmaend))
#define end_dma_cur    (dma_go_cur && HI(ramdmaend))
#define end_dma_vid    (dma_go_vid && HI(ramdmaend))

#define Nactive_high(sig,del) SignalEvt(sig, ONE,state->del-1); else SignalEvt(sig,ZERO,state->del)
#define  active_high(sig,del) SignalEvt(sig, ONE,state->del); else SignalEvt(sig,ZERO,state->del-1)
#define   active_low(sig,del) SignalEvt(sig,ZERO,state->del); else SignalEvt(sig, ONE,state->del-1)

/*============================ DAG control decode block =====================*/

/*----------------------- decode Model Definition----------------------------*/

#define DAGCONC_def  "DAGCONC(prog,dmagoing,lwrite,flybi,vidlast,dagreg,bna_i0, \
  bna_i1,bna_i2,bna_i3,bna_s0,bna_s1,la[],dmach[],ioch[],dmastart)=> \
  (Nwr_c,Nrd_c,Ndr_w2c,Nrd_in,Nrd_co,Nrd_e, \
  dma_go_i0,dma_go_i1,dma_go_i2,dma_go_i3,dma_go_s0,dma_go_s1,dma_govid)"

typedef struct {
  ModelCore model;
  Signal *prog,*dmagoing,*lwrite,*flybi,*vidlast,*dagreg,*bna_i0,*bna_i1,*bna_i2,*bna_i3;
  Signal *bna_s0,*bna_s1;
  Bundle *la,*dmach,*ioch;
  Signal *dmastart;
  Signal *Nwr_c,*Nrd_c,*Ndr_w2c,*Nrd_in,*Nrd_co,*Nrd_e;
  Signal *dma_go_i0,*dma_go_i1,*dma_go_i2,*dma_go_i3,*dma_go_s0,*dma_go_s1,*dma_govid;
} ModelOfDAGCONC;

typedef struct {
  unsigned rd_del,wr_del,dr_del,oe_del;
} DAGCONCState;

/*-------------------- decode Model Code-------------------------------------*/

static void DAGCONC_proc(ModelInst)
VoidStar ModelInst;
{
  ModelOfDAGCONC *me = (ModelOfDAGCONC *) ModelInst;
  DAGCONCState *state = me->model.state;
  unsigned c_reg,in_reg,co_reg,e_reg;
  unsigned la_val,dmach_val,ioch_val;

/* beware of race between driving wbus using prog and strobing data in with prog */

   la_val    = BINVAL(la);
   dmach_val = BINVAL(dmach);
   ioch_val  = BINVAL(ioch);

   c_reg  = ( HI(dagreg) && ( (la_val==0x00) || (la_val==0x02) ||
                              (la_val==0x08) || (la_val==0x0A) ||
                              (la_val==0x10) || (la_val==0x12) ||
                              (la_val==0X18) || (la_val==0x1A) ||
                              (la_val==0x20) || (la_val==0x22) ||
                              (la_val==0x28) || (la_val==0x2A) ||
                              (la_val==0x30) ||
                              (la_val==0x34) ) );

   in_reg = ( HI(dagreg) && ( (0x31==la_val) || (0x37==la_val) ||
                              (0x36==la_val)) );

   co_reg = ( HI(dagreg) && ( (la_val==0x04) ||
                              (la_val==0x0C) ||
                              (la_val==0x14) ||
                              (la_val==0x1C) ||
                              (la_val==0x24) ||
                              (la_val==0x2C) ||
                              (la_val==0x38) ) );

   e_reg  = ( HI(dagreg) && ( (la_val==0x01) || (la_val==0x03) ||
                              (la_val==0x09) || (la_val==0x0B) ||
                              (la_val==0x11) || (la_val==0x13) ||
                              (la_val==0x19) || (la_val==0x1B) ||
                              (la_val==0x21) || (la_val==0x23) ||
                              (la_val==0x29) || (la_val==0x2B) ||
                              (la_val==0x35) ) );

/*-------------- control signals ----------------*/

  if ( HI(prog) && c_reg && LO(lwrite) )          active_low(Nrd_c   ,rd_del);   /*  read current    */
  if ( HI(prog) && c_reg && HI(lwrite) )          active_low(Nwr_c   ,wr_del);   /* write current    */
  if ((LO(dmagoing) && LO(lwrite) && (la_val==0x39) && LO(flybi)) ||            /*  test mode       */
      (LO(prog) && LO(dmastart) && LO(vidlast) && (dmach_val==VID)) ||
      (LO(prog) && LO(dmastart) && (dmach_val!=VID)) ) active_low(Ndr_w2c ,dr_del);   /* write back  */
  if ( HI(prog) && in_reg && LO(lwrite) ||
       read_reg(0x39) )                           active_low(Nrd_in  ,rd_del);   /*  read init       */
  if ( HI(prog) && co_reg && LO(lwrite) )         active_low(Nrd_co  ,rd_del);   /*  read control    */
  if ( HI(prog) &&  e_reg && LO(lwrite) )         active_low(Nrd_e   ,rd_del);   /*  read end        */

  if ( dma_go_i0a || dma_go_i0b )       active_high(dma_go_i0,oe_del);
  if ( dma_go_i1a || dma_go_i1b )       active_high(dma_go_i1,oe_del);
  if ( dma_go_i2a || dma_go_i2b )       active_high(dma_go_i2,oe_del);
  if ( dma_go_i3a || dma_go_i3b )       active_high(dma_go_i3,oe_del);
  if ( dma_go_s0a || dma_go_s0b )       active_high(dma_go_s0,oe_del);
  if ( dma_go_s1a || dma_go_s1b )       active_high(dma_go_s1,oe_del);
  if ( dma_go_vid )                     active_high(dma_govid,oe_del);

}

/*----------------- decode Initialisation-----------------------------*/

static void DAGCONC_init(ModelInst)
VoidStar ModelInst;
{
  DAGCONCState *s;
  ModelOfDAGCONC *me = (ModelOfDAGCONC* ) ModelInst;
  NEW(s, DAGCONCState);

  me->model.proc = DAGCONC_proc;             

  s->rd_del = DELAY("rd_del");
  s->wr_del = DELAY("wr_del");
  s->dr_del = DELAY("dr_del");
  s->oe_del = DELAY("oe_del");

  me->model.state = (VoidStar ) s;

  SignalEvt(    Nwr_c,UNDEF_HIGH,1);
  SignalEvt(    Nrd_c,UNDEF_HIGH,1);
  SignalEvt(  Ndr_w2c,UNDEF_HIGH,1);
  SignalEvt(   Nrd_in,UNDEF_HIGH,1);
  SignalEvt(   Nrd_co,UNDEF_HIGH,1);
  SignalEvt(    Nrd_e,UNDEF_HIGH,1);
  SignalEvt(dma_go_i0,UNDEF_HIGH,1);
  SignalEvt(dma_go_i1,UNDEF_HIGH,1);
  SignalEvt(dma_go_i2,UNDEF_HIGH,1);
  SignalEvt(dma_go_i3,UNDEF_HIGH,1);
  SignalEvt(dma_go_s0,UNDEF_HIGH,1);
  SignalEvt(dma_go_s1,UNDEF_HIGH,1);

}

/*============================ DAG select decode block ======================*/

/*----------------------- decode Model Definition----------------------------*/

#define DAGSELC_def  "DAGSELC(dmagoing,lwrite,dagreg,la[])=> \
  (sel_c_i0,sel_e_i0a,sel_e_i0b,sel_c_i1,sel_e_i1a,sel_e_i1b, \
  sel_c_i2,sel_e_i2a,sel_e_i2b,sel_c_i3,sel_e_i3a,sel_e_i3b, \
  sel_c_s0,sel_e_s0a,sel_e_s0b,sel_c_s1,sel_e_s1a,sel_e_s1b)"

typedef struct {
  ModelCore model;
  Signal *dmagoing,*lwrite,*dagreg;
  Bundle *la;
  Signal *sel_c_i0,*sel_e_i0a,*sel_e_i0b,*sel_c_i1,*sel_e_i1a,*sel_e_i1b;
  Signal *sel_c_i2,*sel_e_i2a,*sel_e_i2b,*sel_c_i3,*sel_e_i3a,*sel_e_i3b;
  Signal *sel_c_s0,*sel_e_s0a,*sel_e_s0b,*sel_c_s1,*sel_e_s1a,*sel_e_s1b;
} ModelOfDAGSELC;

typedef struct {
  unsigned oe_del;
} DAGSELCState;

/*-------------------- decode Model Code-------------------------------------*/

static void DAGSELC_proc(ModelInst)
VoidStar ModelInst;
{
  ModelOfDAGSELC *me = (ModelOfDAGSELC *) ModelInst;
  DAGSELCState *state = me->model.state;
  unsigned la_val;

  la_val    = BINVAL(la);

  if ( sel_w_reg(0x04) )                active_high(sel_c_i0,oe_del);
  if ( sel_w_reg(0x0C) )                active_high(sel_c_i1,oe_del);
  if ( sel_w_reg(0x14) )                active_high(sel_c_i2,oe_del);
  if ( sel_w_reg(0x1C) )                active_high(sel_c_i3,oe_del);
  if ( sel_w_reg(0x24) )                active_high(sel_c_s0,oe_del);
  if ( sel_w_reg(0x2C) )                active_high(sel_c_s1,oe_del);

  if ( sel_w_reg(0x01) )                active_high(sel_e_i0a,oe_del);
  if ( sel_w_reg(0x03) )                active_high(sel_e_i0b,oe_del);
  if ( sel_w_reg(0x09) )                active_high(sel_e_i1a,oe_del);
  if ( sel_w_reg(0x0B) )                active_high(sel_e_i1b,oe_del);
  if ( sel_w_reg(0x11) )                active_high(sel_e_i2a,oe_del);
  if ( sel_w_reg(0x13) )                active_high(sel_e_i2b,oe_del);
  if ( sel_w_reg(0x19) )                active_high(sel_e_i3a,oe_del);
  if ( sel_w_reg(0x1B) )                active_high(sel_e_i3b,oe_del);
  if ( sel_w_reg(0x21) )                active_high(sel_e_s0a,oe_del);
  if ( sel_w_reg(0x23) )                active_high(sel_e_s0b,oe_del);
  if ( sel_w_reg(0x29) )                active_high(sel_e_s1a,oe_del);
  if ( sel_w_reg(0x2B) )                active_high(sel_e_s1b,oe_del);

}

/*----------------- decode Initialisation-----------------------------*/

static void DAGSELC_init(ModelInst)
VoidStar ModelInst;
{
  DAGSELCState *s;
  ModelOfDAGSELC *me = (ModelOfDAGSELC* ) ModelInst;
  NEW(s, DAGSELCState);

  me->model.proc = DAGSELC_proc;             

  s->oe_del = DELAY("oe_del");

  me->model.state = (VoidStar ) s;
}

/*============================ DAG Noe decode block =========================*/

/*----------------------- decode Model Definition----------------------------*/

#define DAGNOEC_def  "DAGNOEC(prog,dmagoing,lwrite,flybi,vidlast,dagreg,bna_i0,\
bna_i1,bna_i2,bna_i3,bna_s0,bna_s1,vncs,la[],dmach[],ioch[],dmastart)=>(N_i0ca,N_i0ea,\
N_i0cb,N_i0eb,N_i0co,N_i0s,N_i1ca,N_i1ea,N_i1cb,N_i1eb,N_i1co,\
N_i1s,N_i2ca,N_i2ea,N_i2cb,N_i2eb,N_i2co,N_i2s,N_i3ca,N_i3ea,\
N_i3cb,N_i3eb,N_i3co,N_i3s,N_s0ca,N_s0ea,N_s0cb,N_s0eb,N_s0co,\
N_s0s,N_s1ca,N_s1ea,N_s1cb,N_s1eb,N_s1co,N_s1s,N_cc,N_cin,N_vc,\
N_ve,oe_ve,N_vst,N_vin,N_vco,N_cco,N_msk,N_sta,N_req,Nnone)"

typedef struct {
  ModelCore model;
  Signal *prog,*dmagoing,*lwrite,*flybi,*vidlast,*dagreg,*bna_i0,*bna_i1,*bna_i2,*bna_i3;
  Signal *bna_s0,*bna_s1,*vncs;
  Bundle *la,*dmach,*ioch;
  Signal *dmastart;
  Signal *N_i0ca,*N_i0ea,*N_i0cb,*N_i0eb,*N_i0co,*N_i0s;
  Signal *N_i1ca,*N_i1ea,*N_i1cb,*N_i1eb,*N_i1co,*N_i1s;
  Signal *N_i2ca,*N_i2ea,*N_i2cb,*N_i2eb,*N_i2co,*N_i2s;
  Signal *N_i3ca,*N_i3ea,*N_i3cb,*N_i3eb,*N_i3co,*N_i3s;
  Signal *N_s0ca,*N_s0ea,*N_s0cb,*N_s0eb,*N_s0co,*N_s0s;
  Signal *N_s1ca,*N_s1ea,*N_s1cb,*N_s1eb,*N_s1co,*N_s1s;
  Signal *N_cc,*N_cin,*N_vc,*N_ve,*oe_ve,*N_vst;
  Signal *N_vin,*N_vco,*N_cco,*N_msk,*N_sta,*N_req,*Nnone;
} ModelOfDAGNOEC;

typedef struct {
  unsigned oe_del;
} DAGNOECState;

/*-------------------- decode Model Code-------------------------------------*/

static void DAGNOEC_proc(ModelInst)
VoidStar ModelInst;
{
  ModelOfDAGNOEC *me = (ModelOfDAGNOEC *) ModelInst;
  DAGNOECState *state = me->model.state;
  unsigned c_reg,co_reg,e_reg;
  unsigned la_val,dmach_val,ioch_val;

/* beware of race between driving wbus using prog and strobing data in with prog */

   la_val    = BINVAL(la);
   dmach_val = BINVAL(dmach);
   ioch_val  = BINVAL(ioch);

   c_reg  = ( HI(dagreg) && ( (la_val==0x00) || (la_val==0x02) ||
                              (la_val==0x08) || (la_val==0x0A) ||
                              (la_val==0x10) || (la_val==0x12) ||
                              (la_val==0X18) || (la_val==0x1A) ||
                              (la_val==0x20) || (la_val==0x22) ||
                              (la_val==0x28) || (la_val==0x2A) ||
                              (la_val==0x30) ||
                              (la_val==0x34) ) );

   co_reg = ( HI(dagreg) && ( (la_val==0x04) ||
                              (la_val==0x0C) ||
                              (la_val==0x14) ||
                              (la_val==0x1C) ||
                              (la_val==0x24) ||
                              (la_val==0x2C) ||
                              (la_val==0x38) ) );

   e_reg  = ( HI(dagreg) && ( (la_val==0x01) || (la_val==0x03) ||
                              (la_val==0x09) || (la_val==0x0B) ||
                              (la_val==0x11) || (la_val==0x13) ||
                              (la_val==0x19) || (la_val==0x1B) ||
                              (la_val==0x21) || (la_val==0x23) ||
                              (la_val==0x29) || (la_val==0x2B) ||
                              (la_val==0x35) ) );

  if  ( read_reg(0x00) || dma_go_i0a )                 active_low(N_i0ca,oe_del);
  if  ( read_reg(0x02) || dma_go_i0b )                 active_low(N_i0cb,oe_del);
  if  ( read_reg(0x08) || dma_go_i1a )                 active_low(N_i1ca,oe_del);
  if  ( read_reg(0x0A) || dma_go_i1b )                 active_low(N_i1cb,oe_del);
  if  ( read_reg(0x10) || dma_go_i2a )                 active_low(N_i2ca,oe_del);
  if  ( read_reg(0x12) || dma_go_i2b )                 active_low(N_i2cb,oe_del);
  if  ( read_reg(0x18) || dma_go_i3a )                 active_low(N_i3ca,oe_del);
  if  ( read_reg(0x1A) || dma_go_i3b )                 active_low(N_i3cb,oe_del);
  if  ( read_reg(0x20) || dma_go_s0a )                 active_low(N_s0ca,oe_del);
  if  ( read_reg(0x22) || dma_go_s0b )                 active_low(N_s0cb,oe_del);
  if  ( read_reg(0x28) || dma_go_s1a )                 active_low(N_s1ca,oe_del);
  if  ( read_reg(0x2A) || dma_go_s1b )                 active_low(N_s1cb,oe_del);
  if  ( read_reg(0x30) || dma_go_cur )                 active_low(N_cc  ,oe_del);
  if  ( read_reg(0x34) || dma_go_vis ||
       ( LO(dmagoing) && !(c_reg && LO(lwrite) && HI(prog))) ) active_low(N_vc  ,oe_del);

  if  ( read_reg(0x01) || dma_go_i0a )                 active_low(N_i0ea,oe_del);
  if  ( read_reg(0x03) || dma_go_i0b )                 active_low(N_i0eb,oe_del);
  if  ( read_reg(0x09) || dma_go_i1a )                 active_low(N_i1ea,oe_del);
  if  ( read_reg(0x0B) || dma_go_i1b )                 active_low(N_i1eb,oe_del);
  if  ( read_reg(0x11) || dma_go_i2a )                 active_low(N_i2ea,oe_del);
  if  ( read_reg(0x13) || dma_go_i2b )                 active_low(N_i2eb,oe_del);
  if  ( read_reg(0x19) || dma_go_i3a )                 active_low(N_i3ea,oe_del);
  if  ( read_reg(0x1B) || dma_go_i3b )                 active_low(N_i3eb,oe_del);
  if  ( read_reg(0x21) || dma_go_s0a )                 active_low(N_s0ea,oe_del);
  if  ( read_reg(0x23) || dma_go_s0b )                 active_low(N_s0eb,oe_del);
  if  ( read_reg(0x29) || dma_go_s1a )                 active_low(N_s1ea,oe_del);
  if  ( read_reg(0x2B) || dma_go_s1b )                 active_low(N_s1eb,oe_del);
  if  ( read_reg(0x35) || dma_go_vic ||
       ( LO(dmagoing) && !(e_reg && LO(lwrite) && HI(prog))) ) active_low(N_ve  ,oe_del);
  if  ( read_reg(0x35) || dma_go_vic ||
       ( LO(dmagoing) && !(e_reg && LO(lwrite) && HI(prog))) ) Nactive_high(oe_ve ,oe_del);

  if  ( read_reg(0x04) || dma_go_i0a || dma_go_i0b )   active_low(N_i0co,oe_del);
  if  ( read_reg(0x0C) || dma_go_i1a || dma_go_i1b )   active_low(N_i1co,oe_del);
  if  ( read_reg(0x14) || dma_go_i2a || dma_go_i2b )   active_low(N_i2co,oe_del);
  if  ( read_reg(0x1C) || dma_go_i3a || dma_go_i3b )   active_low(N_i3co,oe_del);
  if  ( read_reg(0x24) || dma_go_s0a || dma_go_s0b )   active_low(N_s0co,oe_del);
  if  ( read_reg(0x2C) || dma_go_s1a || dma_go_s1b )   active_low(N_s1co,oe_del);
  if  ( !read_reg(0x04) && !dma_go_i0a && !dma_go_i0b &&
        !read_reg(0x0C) && !dma_go_i1a && !dma_go_i1b &&
        !read_reg(0x14) && !dma_go_i2a && !dma_go_i2b &&
        !read_reg(0x1C) && !dma_go_i3a && !dma_go_i3b &&
        !read_reg(0x24) && !dma_go_s0a && !dma_go_s0b &&
        !read_reg(0x2C) && !dma_go_s1a && !dma_go_s1b )   active_low(Nnone,oe_del);

  if  ( read_reg(0x38) || dma_go_vis ||
       ( LO(dmagoing) && !(co_reg && LO(lwrite) && HI(prog))) ) active_low(N_vco ,oe_del);
  if  ( dma_go_cur )                                   active_low(N_cco ,oe_del);

  if  ( read_reg(0x05)   )                             active_low(N_i0s ,oe_del);
  if  ( read_reg(0x0D)   )                             active_low(N_i1s ,oe_del);
  if  ( read_reg(0x15)   )                             active_low(N_i2s ,oe_del);
  if  ( read_reg(0x1D)   )                             active_low(N_i3s ,oe_del);
  if  ( read_reg(0x25)   )                             active_low(N_s0s ,oe_del);
  if  ( read_reg(0x2D)   )                             active_low(N_s1s ,oe_del);

  if (( (!(HI(prog) && HI(dagreg)))  && HI(dmastart) && LO(vncs) && HI(flybi) ) ||                       /* c init -> current  */
      ( read_reg(0x31) ) )                             active_low(N_cin ,oe_del);   /* read cursor init   */
 
  if  ( read_reg(0x36) ||                                                             /* read video start   */
       (LO(dmastart) && (dmach_val==0x4) && HI(vidlast)) ) active_low(N_vst ,oe_del);   /* v start -> current */

  if (( (!(HI(prog) && HI(dagreg))) && HI(dmastart) && HI(vncs) && HI(flybi) ) ||                       /* v init -> current  */
        read_reg(0x37) )                               active_low(N_vin ,oe_del);   /* read init reg      */

  if  ( read_reg(0x3C) )                               active_low(N_sta ,oe_del);
  if  ( read_reg(0x3D) )                               active_low(N_req ,oe_del);
  if  ( read_reg(0x3E) )                               active_low(N_msk ,oe_del);

}

/*----------------- decode Initialisation-----------------------------*/

static void DAGNOEC_init(ModelInst)
VoidStar ModelInst;
{
  DAGNOECState *s;
  ModelOfDAGNOEC *me = (ModelOfDAGNOEC* ) ModelInst;
  NEW(s, DAGNOECState);

  me->model.proc = DAGNOEC_proc;             

  s->oe_del = DELAY("oe_del");

  me->model.state = (VoidStar ) s;

  SignalEvt(N_i0ca,UNDEF_HIGH,1);
  SignalEvt(N_i0ea,UNDEF_HIGH,1);
  SignalEvt(N_i0cb,UNDEF_HIGH,1);
  SignalEvt(N_i0eb,UNDEF_HIGH,1);
  SignalEvt(N_i0co,UNDEF_HIGH,1);
  SignalEvt( N_i0s,UNDEF_HIGH,1);
  SignalEvt(N_i1ca,UNDEF_HIGH,1);
  SignalEvt(N_i1ea,UNDEF_HIGH,1);
  SignalEvt(N_i1cb,UNDEF_HIGH,1);
  SignalEvt(N_i1eb,UNDEF_HIGH,1);
  SignalEvt(N_i1co,UNDEF_HIGH,1);
  SignalEvt( N_i1s,UNDEF_HIGH,1);
  SignalEvt(N_i2ca,UNDEF_HIGH,1);
  SignalEvt(N_i2ea,UNDEF_HIGH,1);
  SignalEvt(N_i2cb,UNDEF_HIGH,1);
  SignalEvt(N_i2eb,UNDEF_HIGH,1);
  SignalEvt(N_i2co,UNDEF_HIGH,1);
  SignalEvt( N_i2s,UNDEF_HIGH,1);
  SignalEvt(N_i3ca,UNDEF_HIGH,1);
  SignalEvt(N_i3ea,UNDEF_HIGH,1);
  SignalEvt(N_i3cb,UNDEF_HIGH,1);
  SignalEvt(N_i3eb,UNDEF_HIGH,1);
  SignalEvt(N_i3co,UNDEF_HIGH,1);
  SignalEvt( N_i3s,UNDEF_HIGH,1);
  SignalEvt(N_s0ca,UNDEF_HIGH,1);
  SignalEvt(N_s0ea,UNDEF_HIGH,1);
  SignalEvt(N_s0cb,UNDEF_HIGH,1);
  SignalEvt(N_s0eb,UNDEF_HIGH,1);
  SignalEvt(N_s0co,UNDEF_HIGH,1);
  SignalEvt( N_s0s,UNDEF_HIGH,1);
  SignalEvt(N_s1ca,UNDEF_HIGH,1);
  SignalEvt(N_s1ea,UNDEF_HIGH,1);
  SignalEvt(N_s1cb,UNDEF_HIGH,1);
  SignalEvt(N_s1eb,UNDEF_HIGH,1);
  SignalEvt(N_s1co,UNDEF_HIGH,1);
  SignalEvt( N_s1s,UNDEF_HIGH,1);
  SignalEvt(  N_cc,UNDEF_HIGH,1);
  SignalEvt( N_cin,UNDEF_HIGH,1);
  SignalEvt(  N_vc,UNDEF_HIGH,1);
  SignalEvt(  N_ve,UNDEF_HIGH,1);
  SignalEvt( N_vst,UNDEF_HIGH,1);
  SignalEvt( N_vin,UNDEF_HIGH,1);
  SignalEvt( N_vco,UNDEF_HIGH,1);
  SignalEvt( N_msk,UNDEF_HIGH,1);
  SignalEvt( N_sta,UNDEF_HIGH,1);
  SignalEvt( N_req,UNDEF_HIGH,1);

}

/*============================ DAG strobe decode block ======================*/

/*----------------------- decode Model Definition----------------------------*/

#define DAGSTC_def  "DAGSTC(prog,dmagoing,ramdmaend,lwrite,flybi,vidlast,dagreg,bna_i0, \
bna_i1,bna_i2,bna_i3,bna_s0,bna_s1,vncs,dmastart,la[],dmach[],ioch[])=>(st_i0ea, \
st_i0eb,st_i0co,st_i1ea,st_i1eb,st_i1co,st_i2ea,st_i2eb, \
st_i2co,st_i3ea,st_i3eb,st_i3co,st_s0ea,st_s0eb,st_s0co, \
st_s1ea,st_s1eb,st_s1co,st_cin,st_ve,st_vst,st_vin,st_vco,st_msk)"

typedef struct {
  ModelCore model;
  Signal *prog,*dmagoing,*ramdmaend,*lwrite,*flybi,*vidlast,*dagreg,*bna_i0,*bna_i1,*bna_i2,*bna_i3;
  Signal *bna_s0,*bna_s1,*vncs,*dmastart;
  Bundle *la,*dmach,*ioch;
  Signal *st_i0ea,*st_i0eb,*st_i0co;
  Signal *st_i1ea,*st_i1eb,*st_i1co;
  Signal *st_i2ea,*st_i2eb,*st_i2co;
  Signal *st_i3ea,*st_i3eb,*st_i3co;
  Signal *st_s0ea,*st_s0eb,*st_s0co;
  Signal *st_s1ea,*st_s1eb,*st_s1co;
  Signal *st_cin,*st_ve,*st_vst,*st_vin,*st_vco,*st_msk;
} ModelOfDAGSTC;

typedef struct {
  unsigned st_del;
} DAGSTCState;

/*-------------------- decode Model Code-------------------------------------*/

static void DAGSTC_proc(ModelInst)
VoidStar ModelInst;
{
  ModelOfDAGSTC *me = (ModelOfDAGSTC *) ModelInst;
  DAGSTCState *state = me->model.state;
  unsigned la_val,dmach_val,ioch_val;

   la_val    = BINVAL(la);
   dmach_val = BINVAL(dmach);
   ioch_val  = BINVAL(ioch);

/*-------------- write latch signals ----------------*/
/* no write-back for current regs as these are sepaerated out */

  if  ( write_reg(0x01)  )                         active_high(st_i0ea,st_del);
  if  ( write_reg(0x03)  )                         active_high(st_i0eb,st_del);
  if  ( write_reg(0x09)  )                         active_high(st_i1ea,st_del);
  if  ( write_reg(0x0B)  )                         active_high(st_i1eb,st_del);
  if  ( write_reg(0x11)  )                         active_high(st_i2ea,st_del);
  if  ( write_reg(0x13)  )                         active_high(st_i2eb,st_del);
  if  ( write_reg(0x19)  )                         active_high(st_i3ea,st_del);
  if  ( write_reg(0x1B)  )                         active_high(st_i3eb,st_del);
  if  ( write_reg(0x21)  )                         active_high(st_s0ea,st_del);
  if  ( write_reg(0x23)  )                         active_high(st_s0eb,st_del);
  if  ( write_reg(0x29)  )                         active_high(st_s1ea,st_del);
  if  ( write_reg(0x2B)  )                         active_high(st_s1eb,st_del);
  if  ( write_reg(0x35)  )                         active_high(st_ve  ,st_del);
  if  ( write_reg(0x38)  )                         active_high(st_vco ,st_del);

  if  ( write_reg(0x04)  )                         active_high(st_i0co,st_del);
  if  ( write_reg(0x0C)  )                         active_high(st_i1co,st_del);
  if  ( write_reg(0x14)  )                         active_high(st_i2co,st_del);
  if  ( write_reg(0x1C)  )                         active_high(st_i3co,st_del);
  if  ( write_reg(0x24)  )                         active_high(st_s0co,st_del);
  if  ( write_reg(0x2C)  )                         active_high(st_s1co,st_del);

  if  ( write_reg(0x31)  )                         active_high(st_cin ,st_del);
  if  ( write_reg(0x36)  )                         active_high(st_vst ,st_del);
  if  ( write_reg(0x37)  )                         active_high(st_vin ,st_del);

  if  ( write_reg(0x3E)  )                         active_high(st_msk ,st_del);
}

/*----------------- decode Initialisation-----------------------------*/

static void DAGSTC_init(ModelInst)
VoidStar ModelInst;
{
  DAGSTCState *s;
  ModelOfDAGSTC *me = (ModelOfDAGSTC* ) ModelInst;
  NEW(s, DAGSTCState);

  me->model.proc = DAGSTC_proc;             

  s->st_del = DELAY("st_del");

  me->model.state = (VoidStar ) s;
}

/*============================ DAG write back strobe decode block ===========*/

/*----------------------- decode Model Definition----------------------------*/

#define DAGSBC_def  "DAGSBC(prog,dmagoing,ramdmaend,lwrite,flybi,vidlast,dagreg,bna_i0, \
bna_i1,bna_i2,bna_i3,bna_s0,bna_s1,vncs,dmastart,la[],dmach[],ioch[])=>(sb_i0ca,sb_i0cb, \
sb_i1ca,sb_i1cb,sb_i2ca,sb_i2cb,sb_i3ca,sb_i3cb,sb_s0ca,sb_s0cb,sb_s1ca,sb_s1cb,sb_cc,sb_vc, \
sd_vc,ck_i0a,ck_i0b,ck_i1a,ck_i1b,ck_i2a,ck_i2b,ck_i3a,ck_i3b,ck_s0a,ck_s0b,ck_s1a,ck_s1b)"

typedef struct {
  ModelCore model;
  Signal *prog,*dmagoing,*ramdmaend,*lwrite,*flybi,*vidlast,*dagreg,*bna_i0,*bna_i1,*bna_i2,*bna_i3;
  Signal *bna_s0,*bna_s1,*vncs,*dmastart;
  Bundle *la,*dmach,*ioch;
  Signal *sb_i0ca,*sb_i0cb;
  Signal *sb_i1ca,*sb_i1cb;
  Signal *sb_i2ca,*sb_i2cb;
  Signal *sb_i3ca,*sb_i3cb;
  Signal *sb_s0ca,*sb_s0cb;
  Signal *sb_s1ca,*sb_s1cb;
  Signal *sb_cc,*sb_vc,*sd_vc;
  Signal *ck_i0a,*ck_i0b,*ck_i1a,*ck_i1b;
  Signal *ck_i2a,*ck_i2b,*ck_i3a,*ck_i3b;
  Signal *ck_s0a,*ck_s0b,*ck_s1a,*ck_s1b;
} ModelOfDAGSBC;

typedef struct {
  unsigned st_del;
} DAGSBCstate;

/*-------------------- decode Model Code-------------------------------------*/

static void DAGSBC_proc(ModelInst)
VoidStar ModelInst;
{
  ModelOfDAGSBC *me = (ModelOfDAGSBC *) ModelInst;
  DAGSBCstate *state = me->model.state;
  unsigned la_val,dmach_val,ioch_val;

   la_val    = BINVAL(la);
   dmach_val = BINVAL(dmach);
   ioch_val  = BINVAL(ioch);

/*-------------- write latch signals ----------------*/

  if  ( write_reg(0x00) || end_dma_i0a )           active_high(sb_i0ca,st_del);
  if  ( write_reg(0x08) || end_dma_i1a )           active_high(sb_i1ca,st_del);
  if  ( write_reg(0x02) || end_dma_i0b )           active_high(sb_i0cb,st_del);
  if  ( write_reg(0x0A) || end_dma_i1b )           active_high(sb_i1cb,st_del);
  if  ( write_reg(0x10) || end_dma_i2a )           active_high(sb_i2ca,st_del);
  if  ( write_reg(0x12) || end_dma_i2b )           active_high(sb_i2cb,st_del);
  if  ( write_reg(0x18) || end_dma_i3a )           active_high(sb_i3ca,st_del);
  if  ( write_reg(0x1A) || end_dma_i3b )           active_high(sb_i3cb,st_del);
  if  ( write_reg(0x20) || end_dma_s0a )           active_high(sb_s0ca,st_del);
  if  ( write_reg(0x22) || end_dma_s0b )           active_high(sb_s0cb,st_del);
  if  ( write_reg(0x28) || end_dma_s1a )           active_high(sb_s1ca,st_del);
  if  ( write_reg(0x2A) || end_dma_s1b )           active_high(sb_s1cb,st_del);
  if  ( write_reg(0x30) || end_dma_cur ||
      ( (!(HI(prog) && HI(dagreg))) && HI(dmastart) && LO(vncs) && HI(flybi) ))   active_high(sb_cc  ,st_del);
  if  ( write_reg(0x34) || end_dma_vid ||
      ( (!(HI(prog) && HI(dagreg))) && HI(dmastart) && HI(vncs) && HI(flybi) ))   active_high(sb_vc  ,st_del);
  if  ( end_dma_vid ||
      ( (!(HI(prog) && HI(dagreg))) && HI(dmastart) && HI(vncs) && HI(flybi) ))   active_low(sd_vc  ,st_del);

  if  ( write_reg(0x01) || end_dma_i0a )           active_low(ck_i0a,st_del);
  if  ( write_reg(0x03) || end_dma_i0b )           active_low(ck_i0b,st_del);
  if  ( write_reg(0x09) || end_dma_i1a )           active_low(ck_i1a,st_del);
  if  ( write_reg(0x0B) || end_dma_i1b )           active_low(ck_i1b,st_del);
  if  ( write_reg(0x11) || end_dma_i2a )           active_low(ck_i2a,st_del);
  if  ( write_reg(0x13) || end_dma_i2b )           active_low(ck_i2b,st_del);
  if  ( write_reg(0x19) || end_dma_i3a )           active_low(ck_i3a,st_del);
  if  ( write_reg(0x1B) || end_dma_i3b )           active_low(ck_i3b,st_del);
  if  ( write_reg(0x21) || end_dma_s0a )           active_low(ck_s0a,st_del);
  if  ( write_reg(0x23) || end_dma_s0b )           active_low(ck_s0b,st_del);
  if  ( write_reg(0x29) || end_dma_s1a )           active_low(ck_s1a,st_del);
  if  ( write_reg(0x2B) || end_dma_s1b )           active_low(ck_s1b,st_del);

}

/*----------------- decode Initialisation-----------------------------*/

static void DAGSBC_init(ModelInst)
VoidStar ModelInst;
{
  DAGSBCstate *s;
  ModelOfDAGSBC *me = (ModelOfDAGSBC *) ModelInst;
  NEW(s, DAGSBCstate);

  me->model.proc = DAGSBC_proc;             

  s->st_del = DELAY("st_del");

  me->model.state = (VoidStar ) s;

  SignalEvt(ck_i0a,UNDEF_HIGH,1);
  SignalEvt(ck_i0b,UNDEF_HIGH,1);
  SignalEvt(ck_i1a,UNDEF_HIGH,1);
  SignalEvt(ck_i1b,UNDEF_HIGH,1);
  SignalEvt(ck_i2a,UNDEF_HIGH,1);
  SignalEvt(ck_i2b,UNDEF_HIGH,1);
  SignalEvt(ck_i3a,UNDEF_HIGH,1);
  SignalEvt(ck_i3b,UNDEF_HIGH,1);
  SignalEvt(ck_s0a,UNDEF_HIGH,1);
  SignalEvt(ck_s0b,UNDEF_HIGH,1);
  SignalEvt(ck_s1a,UNDEF_HIGH,1);
  SignalEvt(ck_s1b,UNDEF_HIGH,1);
}

/*============================ buffer state machine ==========================*/

#define IDLE_A 6  /* Idle and waiting for buffer A to be written              */
#define IDLE_B 7  /* Idle and waiting for buffer B to be written              */
#define BUSI_A 2  /* Buffer A active, interrupting for buffer B to be written */
#define BUSI_B 3  /* Buffer B active, interrupting for buffer A to be written */
#define BUSY_A 0  /* Buffer A active, buffer B ready (written to)             */
#define BUSY_B 1  /* Buffer B active, buffer A ready                          */

#define DELY 1500

#define WRITE_CLEAR (              HI(clr_addr) && HI(d7))
#define WRITE_BUFA  (HI(enable) && HI(enda_addr))
#define WRITE_BUFB  (HI(enable) && HI(endb_addr))
#define LASTA        (HI(lasta) && HI(dmago))
#define LASTB        (HI(lastb) && HI(dmago))

/*----------------------- Buffer sm Model Definition----------------------------*/
/*  clr_addr = d[7] & lwrite & (la[7:2]==iii000) where iii=i/o ch decoded */
/*  enda_addr =       lwrite & (la[7:2]==iii001)                          */
/*  endb_addr =       lwrite & (la[7:2]==iii011)                          */
/**************************************************************************/

#define DAGBSMC_def  "DAGBSMC(clk,reset,clr_addr,d7,enda_addr,endb_addr,dmago,lasta,lastb,stopa,stopb,enable)=>(q[])"

typedef struct {
  ModelCore model;
  Signal *clk,*reset,*clr_addr,*d7,*enda_addr,*endb_addr,*dmago;
  Signal *lasta,*lastb,*stopa,*stopb,*enable;
  Bundle *q;
} ModelOfDAGBSMC;

typedef struct {
  unsigned old_state;
} DAGBSMCState;

/*-------------------- Buffer sm Model Code-------------------------------------*/

#define new_state(next) {BundleEvt(q,next,DELY); state->old_state=next; break;}

static void DAGBSMC_proc(ModelInst)
VoidStar ModelInst;
{
  ModelOfDAGBSMC *me = (ModelOfDAGBSMC *) ModelInst;
  DAGBSMCState *state = me->model.state;

  if (RISING(clk)) {
     if (HI(reset) || WRITE_CLEAR) {
        BundleEvt(q, IDLE_A, DELY);
        state->old_state=IDLE_A;
     }
     else {
        switch(state->old_state) {
           case IDLE_A: if      (WRITE_BUFA)           new_state(BUSI_A)
                        else                           break;
           case IDLE_B: if      (WRITE_BUFB)           new_state(BUSI_B)
                        else                           break;
           case BUSI_A: if      (WRITE_BUFB)           new_state(BUSY_A)
                        else if (LASTA)                new_state(IDLE_A)
                        else                           break;
           case BUSI_B: if      (WRITE_BUFA)           new_state(BUSY_B)
                        else if (LASTB)                new_state(IDLE_B)
                        else                           break;
           case BUSY_A: if      (LASTA && (HI(stopa))) new_state(IDLE_B)
                        else if (LASTA && (LO(stopa))) new_state(BUSI_B)
                        else                           break;
           case BUSY_B: if      (LASTB && (HI(stopb))) new_state(IDLE_A)
                        else if (LASTB && (LO(stopb))) new_state(BUSI_A)
                        else                           break;
        }
     }
  }
}

/*----------------- Buffer sm Initialisation-----------------------------*/

static void DAGBSMC_init(ModelInst)
VoidStar ModelInst;
{
  DAGBSMCState *s;
  ModelOfDAGBSMC *me = (ModelOfDAGBSMC* ) ModelInst;
  NEW(s, DAGBSMCState);

  me->model.proc = DAGBSMC_proc;             

  BundleEvt(q, IDLE_A, DELY);
  s->old_state=IDLE_A;

  me->model.state = (VoidStar ) s;
}

/*----------------------- incrementer Model Definition----------------------------*/

#define DAGINCC_def  "DAGINCC(in[],inch[],incl[])=>(out[])"

typedef struct {
  ModelCore model;
  Bundle *in,*inch,*incl,*out;
} ModelOfDAGINCC;

typedef struct {
  unsigned inc_del;
} DAGINCCState;

/*-------------------- Buffer sm Model Code-------------------------------------*/

static void DAGINCC_proc(ModelInst)
VoidStar ModelInst;
{
  ModelOfDAGINCC *me = (ModelOfDAGINCC *) ModelInst;
  DAGINCCState *state = me->model.state;
  unsigned in_val,inch_val,incl_val;

  in_val=BINVAL(in);
  inch_val=BINVAL(inch);
  incl_val=BINVAL(incl);

  BundleEvt(out, in_val + incl_val + (inch_val<<8) , state->inc_del);

}

/*----------------- Buffer sm Initialisation-----------------------------*/

static void DAGINCC_init(ModelInst)
VoidStar ModelInst;
{
  DAGINCCState *s;
  ModelOfDAGINCC *me = (ModelOfDAGINCC* ) ModelInst;
  NEW(s, DAGINCCState);

  me->model.proc = DAGINCC_proc;             

  s->inc_del = DELAY("");

  me->model.state = (VoidStar ) s;
}

#define PAUSE_ADDRESS     3
#define ACTIVE_ADDRESS    2
#define INACTIVE_ADDRESS  1
#define COUNT_ADDRESS     0

/*============================ SIGGENC control decode block =====================*/
/* in should be tied to out in the sidle.                                        */
/*                                                                               */
/* The output signal (out) can be either active high or active low and is set in */
/* the sidle using (active_state="high") or (active_state="low")                 */
/* There are four variables "count" "active" "inactive" "pause" which are set    */
/* initially by the sidle. There are two modes of operation, depanding on wether */
/* "active" is zero or not. If "active" is zero then the output signal will go   */
/* to its active state and stay there until the input signal "clr" goes to the   */
/* active state. If "active" is non zero then the output will go to its active   */
/* state for the time set by "active". The variable "count" sets the number of   */
/* active transitions to be generated. A value of zero effectively disables the  */
/* block. For a value of 1 the time "pause" will elapse between setting "count"  */
/* and a single active transition. For "count" greater than 1 the first active   */
/* transition will occur after time "pause". Then, when the active state is      */
/* cleared (either by the signal "clr" if active==0 or after time "active" if    */
/* if active!=0) the next active transition will occur after time "inactive".    */
/* The variables can be programmed to new values using Nprog,address and data.   */
/*                                                                               */
/* examples: 1) A continuous square wave starting after 5us, with period 100ns   */
/*                                                                               */
/* sqg = SIGGENC(square,Vss,address[3:2],Vdd,Vdd,data[31:0]) => square           */
/*    WITH (active_state="high",count=-1,active=ns_50,inactive=ns_50,pause=us_5) */
/*                                                                               */
/* 2) A req/ack requesting block under program control                           */
/*                                                                               */
/* rap = SIGGENC(Nreq,Nack,address[3:2],Nprog,Nrw,data[31:0] => Nreq             */
/*    WITH (active_state="low",count=0,active=0,inactive=ns_250,pause=ns_100     */
/* then to generate, say 3 requests, program count=3                             */
/*                                                                               */
/* NEW mode: if active=0, inactive=0 => simple PIO output with delay "pause".    */
/*  - program non-zero into count for high, zero into count for low              */
/*                                                                               */
/*===============================================================================*/


/*----------------------- decode Model Definition--------------------------------*/

#define SIGGENC_def  "SIGGENC(in,clr,address[],Nprog,Nrw,data[])=>out"

typedef struct {
  ModelCore model;
  Signal *in,*clr;
  Bundle *address;
  Signal *Nprog,*Nrw;
  Bundle *data;
  Signal *out;
} ModelOfSIGGENC;

typedef struct {
  unsigned pause,active,inactive,count,active_state,inactive_state;
} SIGGENCState;

/*-------------------- decode Model Code-------------------------------------*/

static void SIGGENC_proc(ModelInst)
VoidStar ModelInst;
{
  ModelOfSIGGENC *me = (ModelOfSIGGENC *) ModelInst;
  SIGGENCState *state = me->model.state;
  unsigned address_val,data_val;

  if (FALLING(Nprog)) {
     if (LO(Nrw)) {
        address_val = BINVAL(address);
        if      (address_val==   COUNT_ADDRESS) BundleEvt(data,state->count,3000);
        else if (address_val==   PAUSE_ADDRESS) BundleEvt(data,state->pause,3000);
        else if (address_val==  ACTIVE_ADDRESS) BundleEvt(data,state->active,3000);
        else if (address_val==INACTIVE_ADDRESS) BundleEvt(data,state->inactive,3000);
     }
  }

  if (RISING(Nprog)) {
     address_val = BINVAL(address);
     data_val    = BINVAL(data);
     if (HI(Nrw)) {
        if      (address_val==   PAUSE_ADDRESS)      state->pause    = data_val;
        else if (address_val==  ACTIVE_ADDRESS)      state->active   = data_val;
        else if (address_val==INACTIVE_ADDRESS)      state->inactive = data_val;
        else if (address_val==   COUNT_ADDRESS) {
           state->count = data_val;
           if (state->count>0) {
              if ((state->active==0) && (state->inactive==0)) {
                 SignalEvt(out,ONE,state->pause);
              }
              else SignalEvt(out,state->active_state,state->pause);
              state->count--;
           }
           else if ((state->active==0) && (state->inactive==0)) {
              SignalEvt(out,ZERO,state->pause);
           }
        }
     }
     else if (LO(Nrw)) {
        BundleZZZ(data,200);
     }
  }

  if (CHANGING(in)) {
     if ((state->active==0) && (state->inactive==0)) ;
     else {
        if (me->in->value==state->active_state) {
           if (state->active!=0) SignalEvt(out,state->inactive_state,state->active);
        }
        else {
           if (state->count>0) {
              SignalEvt(out,state->active_state,state->inactive);
              state->count--;
           }
        }
     }
  }
       
  if (CHANGING(clr) && (state->active==0) && (state->inactive!=0)) {
     if (me->clr->value==state->active_state) SignalEvt(out,state->inactive_state,1);
  }
}

/*----------------- decode Initialisation-----------------------------*/

static void SIGGENC_init(ModelInst)
VoidStar ModelInst;
{
  SIGGENCState *s;
  ModelOfSIGGENC *me = (ModelOfSIGGENC* ) ModelInst;
  Parameter param;
  NEW(s, SIGGENCState);

  me->model.proc = SIGGENC_proc;             

  s->pause     = DELAY("pause");
  s->active    = DELAY("active");
  s->inactive  = DELAY("inactive");

  if (s->active==1) s->active=0;
  if (s->inactive==1) s->inactive=0;

  param = simmodel_param(me->model.state, "count");
  if (simmodel_type(param) != PT_INTEGER) errhandler_f1(SS_MODELS+DISASTER,
       "%s: SIGGENC 'count' is NOT properly specified\n", (int)INSTNAME);
  s->count = simmodel_int(param);

  param = simmodel_param(me->model.state, "active_state");
  if      (cistrcmp(param,"high")) {
     s->active_state   = ONE;
     s->inactive_state = ZERO;
  }
  else if (cistrcmp(param,"low")) {
     s->active_state   = ZERO;
     s->inactive_state = ONE;
  }
  else errhandler_f1(SS_MODELS+DISASTER,
       "%s: SIGGENC 'active_state' is NOT properly specified (high or low)\n", (int)INSTNAME);

  me->model.state = (VoidStar ) s;

  SignalEvt(out,s->inactive_state,1);

}


/*****************************Model Declaration****************************/

void DAGblks_declare()
{
  simmodel_begin("iomd.dagblks");
  simmodel_declare_s(DAGBSMC_def, DAGBSMC_init);
  simmodel_declare_s(DAGCONC_def, DAGCONC_init);
  simmodel_declare_s(DAGSELC_def, DAGSELC_init);
  simmodel_declare_s(DAGNOEC_def, DAGNOEC_init);
  simmodel_declare_s(DAGSTC_def , DAGSTC_init) ;
  simmodel_declare_s(DAGSBC_def , DAGSBC_init) ;
  simmodel_declare_s(DAGINCC_def, DAGINCC_init);
  simmodel_declare_s(SIGGENC_def, SIGGENC_init);
  simmodel_end();
}
