/* ioint.c */
/* ############################################################### */
/* ############################################################### */
/* ###                                                         ### */
/* ###  IOINT block level model                                ### */
/* ###                                                         ### */
/* ###  Created 5/11/92: Simon Segars                          ### */
/* ###                                                         ### */
/* ###                                                         ### */
/* ############################################################### */

#include <Handy.h>
#include "ioint.h"

/* ############################################################### */

typedef struct
{
  ModelCore     model;
  Signal        *dummy;

} *DelayModel;

static int get_delay (delay_name, model)
char *delay_name;
VoidStar model;
{
  Parameter    param;
  int          delay = 1000*NS;
  DelayModel   me = (DelayModel)model;

  param = simmodel_param(me->model.state,"delay");
  if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  if (simmodel_type(param) == PT_PARAMBLOCK) {
    param = simmodel_sub_param(param,delay_name);
    if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  }
  errhandler_f (SS_MODELS+DISASTER,"!!! %s delay not correctly specified\n",delay_name);
  return delay;
}

#define IOINT_def \
"IOINT(data[7:0], la[6:2], clk2, clk32, dagint, dec_conr, dec_fiqsr, dec_fiqrr, dec_fiqmr,\
       dec_irqasr, dec_irqarr, dec_irqamr, dec_irqbsr, dec_irqbrr, dec_irqbmr, fdrq, flybi, iiccin, iicdin, idin,\
       iointreg, irqa, irqb, kbrx, kbtx, latchtx0, latchtx1, load0, load1, loadtx0, loadtx1, lwrite, Nfintr, Nfiqin,\
       Niext, Nindex, Npfiq, Npirq, Nscirq, Nsintr, pintr, por, prog, rclk, ref8m, reset, testbus, testclk, testmux,\
       timclk, tim0rd, tim1rd, txch0, txch1, txcl0, txcl1, txll0, txll1, txlh0, txlh1, uftx0, uftx1)\
      => (time0[15:0], time1[15:0], iiccout, iicdout, idout, kcon_rd, kcon_wt, kdata_rd, kdata_wt, Neniicc, Neniicd,\
          Nenid, Ninitu, Nfiq, Nirq)"

typedef struct {
    ModelCore   model;
    Bundle      *data, *la;
    Signal      *clk2, *clk32, *dagint, *dec_conr, *dec_fiqsr, *dec_fiqrr, *dec_fiqmr;
    Signal      *dec_irqasr, *dec_irqarr, *dec_irqamr, *dec_irqbsr, *dec_irqbrr, *dec_irqbmr;
    Signal      *fdrq, *flybi, *iiccin, *iicdin, *idin, *iointreg, *irqa, *irqb, *kbrx, *kbtx;
    Signal      *latchtx0, *latchtx1, *load0, *load1, *loadtx0, *loadtx1, *lwrite;
    Signal      *Nfintr, *Nfiqin, *Niext, *Nindex, *Npfiq, *Npirq, *Nscirq, *Nsintr, *pintr, *por, *prog;
    Signal      *rclk, *ref8m, *reset, *testbus, *testclk, *testmux, *timclk;
    Signal      *tim0rd, *tim1rd, *txch0, *txch1, *txcl0, *txcl1, *txll0, *txll1, *txlh0, *txlh1, *uftx0, *uftx1;
    Bundle      *time0, *time1;
    Signal      *iiccout, *iicdout, *idout, *kcon_rd, *kcon_wt, *kdata_rd, *kdata_wt, *Neniicc, *Neniicd;
    Signal      *Nenid, *Ninitu, *Nfiq, *Nirq;
    } *IOINT_model;

typedef struct {
    int dt_lch, dt_oe, dt_gt;
    int time0, latch0h, latch0l, count0h, count0l, fload0, underflow0;
    int time1, latch1h, latch1l, count1h, count1l, fload1, underflow1;
    int tclk, tprog;
    int control, fiq, fiql, fmask;
    int irqa, irqau, irqal, imaska;
    int irqb, irqbl, imaskb;
    int dec_irqas, dec_irqar, dec_irqam;
    int dec_irqbs, dec_irqbr, dec_irqbm;
    int dec_fiqs, dec_fiqr, dec_fiqm, dec_con;
    int pintrl, nindexl, nindexs;
    int idinl, idins, iiccinl, iiccins, iicdinl, iicdins, flybics;
    } IOINT_state;

/*--------------------------------------------------------------------------*/

static void IOINT_proc (modelInst)
VoidStar modelInst;
{
    IOINT_model me = (IOINT_model) modelInst;
    IOINT_state *s = (IOINT_state *)me->model.state;
    int add, enable, kcon, kdat;

/*-----------------------Ninitu----------------------------------------------*/

  SignalEvt(Ninitu, (LO(por) || HI(Nfiqin)), s->dt_gt);

/*--------------------------Address Decoder----------------------------------*/

  add = BINVAL(la);

  /* Control Register Decode */

  s->dec_con = (HI(prog) && (add==0x00) && HI(lwrite) && HI(iointreg));

     SignalEvt(dec_conr, ((add==0x00) && LO(lwrite) && HI(iointreg) && HI(prog)), s->dt_gt);

  /* Keyboard Decodes */

  kdat = HI(iointreg) && (add==0x01);
  kcon = HI(iointreg) && (add==0x02);

  SignalEvt(kdata_wt, (HI(prog) && HI(lwrite) && kdat), s->dt_gt);
  SignalEvt(kcon_wt,  (HI(prog) && HI(lwrite) && kcon), s->dt_gt);

     SignalEvt(kdata_rd, (kdat && LO(lwrite) && HI(prog)), s->dt_lch);
     SignalEvt(kcon_rd,  (kcon && LO(lwrite) && HI(prog)), s->dt_lch);

  /* IRQA Decodes */

  s->dec_irqas = ((add==0x04) && HI(iointreg)); /* IRQ_A Status register   0x10 */
  s->dec_irqar = ((add==0x05) && HI(iointreg)); /* IRQ_A Request register  0x14 */
  s->dec_irqam = ((add==0x06) && HI(iointreg)); /* IRQ_A Mask register     0x18 */

     SignalEvt(dec_irqasr, (s->dec_irqas && LO(lwrite) && HI(prog)), s->dt_gt); /* IRQ_A Status strobe */
     SignalEvt(dec_irqarr, (s->dec_irqar && LO(lwrite) && HI(prog)), s->dt_gt); /* IRQ_A Request strobe */
     SignalEvt(dec_irqamr, (s->dec_irqam && LO(lwrite) && HI(prog)), s->dt_gt); /* IRQ_A Mask strobe */

  /* IRQB Decodes */

  s->dec_irqbs = ((add==0x08) && HI(iointreg)); /* IRQ_B Status register   0x20 */
  s->dec_irqbr = ((add==0x09) && HI(iointreg)); /* IRQ_B Request register  0x24 */
  s->dec_irqbm = ((add==0x0A) && HI(iointreg)); /* IRQ_B Mask register     0x28 */

     SignalEvt(dec_irqbsr, (s->dec_irqbs && LO(lwrite) && HI(prog)), s->dt_gt); /* IRQ_B Status strobe */
     SignalEvt(dec_irqbrr, (s->dec_irqbr && LO(lwrite) && HI(prog)), s->dt_gt); /* IRQ_B Request strobe */
     SignalEvt(dec_irqbmr, (s->dec_irqbm && LO(lwrite) && HI(prog)), s->dt_gt); /* IRQ_B Mask strobe */

  /* FIQ Decodes */

  s->dec_fiqs = ((add==0x0C) && HI(iointreg)); /* FIQ Status register   0x30 */
  s->dec_fiqr = ((add==0x0D) && HI(iointreg)); /* FIQ Request register  0x34 */
  s->dec_fiqm = ((add==0x0E) && HI(iointreg)); /* FIQ Mask register     0x38 */

     SignalEvt(dec_fiqsr, (s->dec_fiqs && LO(lwrite) && HI(prog)), s->dt_gt); /* FIQ Status strobe */
     SignalEvt(dec_fiqrr, (s->dec_fiqr && LO(lwrite) && HI(prog)), s->dt_gt); /* FIQ Request strobe */
     SignalEvt(dec_fiqmr, (s->dec_fiqm && LO(lwrite) && HI(prog)), s->dt_gt); /* FIQ Mask strobe */

/*-------------------------- Control Register --------------------------------*/

  if      HI(por)   SignalEvt(Nenid, FALSE, s->dt_gt);
  else if HI(reset) SignalEvt(Nenid,  TRUE, s->dt_gt);

  if HI(reset) {
     SignalEvt(Nenid,   TRUE, s->dt_gt);
     SignalEvt(Neniicc, TRUE,  s->dt_gt);
     SignalEvt(Neniicd, TRUE,  s->dt_gt);
   }

  if FALLING(por)
     SignalEvt(Nenid, TRUE, s->dt_gt);

  /* Control Register Read */

   if HI(dec_conr)
      BundleEvt(data, (((s->flybics&0x1)<<7) | 
                       ((s->nindexs&0x1)<<6) |
                       ((s->idins&0x1)  <<3) |
                       ((s->iiccins&0x1)<<1) |
                        (s->iicdins&0x1)     |
                        0x34), s->dt_gt);

   if FALLING(dec_conr)
      BundleZZZ(data, s->dt_oe);

  /* Control Register Writes */

   if (s->dec_con) {
      SignalEvt(Nenid,   ((BINVAL(data)&0x08)>>3), s->dt_gt);
      SignalEvt(Neniicc, ((BINVAL(data)&0x02)>>1), s->dt_gt);
      SignalEvt(Neniicd,  (BINVAL(data)&0x01),     s->dt_gt);
      }

  /* Control Register outputs */

   if CHANGING(testbus) {
      SignalEvt(idout,   HI(testbus), s->dt_gt);
      SignalEvt(iicdout, HI(testbus), s->dt_gt);
      SignalEvt(iiccout, HI(testbus), s->dt_gt);
      }

  /* Synchronise Nindex again, just for the level read in this register */
  /* Also synchronise iicin, iicdin and idin */

   if RISING(rclk) {
      s->nindexl = HI(Nindex);
      s->idinl = HI(idin);
      s->iiccinl = HI(iiccin);
      s->iicdinl = HI(iicdin);
      }
 
   if LO(rclk) {
      s->flybics = HI(flybi);
      s->nindexs = s->nindexl;
      s->idins = s->idinl;
      s->iiccins = s->iiccinl;
      s->iicdins = s->iicdinl;
      }

/*--------------------------IOC Timer interface------------------------------*/

  if (HI(prog) && HI(lwrite) && HI(iointreg))
     s->tprog = TRUE;
  if FALLING(rclk)
     s->tprog = FALSE;

  enable = (HI(rclk) && LO(clk32) && s->tprog);

  /* Timer 0 write enables */

  SignalEvt(txll0,    (enable && (add==0x10)), s->dt_gt);
  SignalEvt(txlh0,    (enable && (add==0x11)), s->dt_gt);
  SignalEvt(loadtx0,  (enable && (add==0x12)), s->dt_gt);
  SignalEvt(latchtx0, (enable && (add==0x13)), s->dt_gt);

  /* Timer 1 write enables */

  SignalEvt(txll1,    (enable && (add==0x14)), s->dt_gt);
  SignalEvt(txlh1,    (enable && (add==0x15)), s->dt_gt);
  SignalEvt(loadtx1,  (enable && (add==0x16)), s->dt_gt);
  SignalEvt(latchtx1, (enable && (add==0x17)), s->dt_gt);

  /* Read strobe for timer 0 */

  SignalEvt(tim0rd, (LO(lwrite) && HI(prog) && HI(iointreg) && ((add&0xFE)==0x10)), s->dt_gt);

  if FALLING(tim0rd)
     BundleZZZ(data, s->dt_oe);

  SignalEvt(txch0, (HI(tim0rd) && (add==0x11)), s->dt_gt);
  SignalEvt(txcl0, (HI(tim0rd) && (add==0x10)), s->dt_gt);

  /* Read strobe for timer 1 */

  SignalEvt(tim1rd, (LO(lwrite) && HI(prog) && HI(iointreg) && ((add&0x14)==0x14)), s->dt_gt);

  if FALLING(tim1rd)
     BundleZZZ(data, s->dt_oe);

  SignalEvt(txch1, (HI(tim1rd) && ((add&0x01)==0x01)), s->dt_gt);
  SignalEvt(txcl1, (HI(tim1rd) && ((add&0x01)==0x00)), s->dt_gt);

  /* Clock Multiplexor */

  s->tclk = LO(testmux) ? HI(clk2) : HI(testclk);

  if RISING(clk32) /* Retime to clk32 to be on the safe side */
     SignalEvt(timclk, s->tclk, s->dt_lch);

/*-------------------------------Timer 0 ------------------------------------*/

 /* The timer itself */
 
  if HI(load0)
     s->time0 = (s->latch0h<<8 | s->latch0l);
  else if RISING(timclk)
     s->time0 = (s->time0 - 1) & 0xFFFF;
 
  BundleEvt(time0, s->time0, s->dt_gt);
 
  /* Reload logic - SR then D-type */
 
  s->underflow0 = (s->time0==0);
 
  if HI(loadtx0)
     s->fload0 = TRUE;
  if HI(load0)
     s->fload0 = FALSE;
 
  if FALLING(timclk)
     SignalEvt(load0, (s->fload0 || s->underflow0), s->dt_lch);
 
  /* Write Latches */
 
  if HI(txll0)
     s->latch0l = BINVAL(data);
 
  if HI(txlh0)
     s->latch0h = BINVAL(data);
 
  /* Count latches */
 
  if HI(latchtx0) {
     s->count0h = (s->time0>>8)&0xFF;
     s->count0l = (s->time0)&0xFF;
     }
 
  /* Count reads */
 
  if HI(txch0)
    BundleEvt(data, s->count0h, s->dt_oe);
  if HI(txcl0)
    BundleEvt(data, s->count0l, s->dt_oe);
 
  /* Interupt output */
 
  if FALLING(timclk)
     SignalEvt(uftx0, s->underflow0, s->dt_gt);

/*-------------------------------Timer 1 ------------------------------------*/

 /* The timer itself */

  if HI(load1)
     s->time1 = (s->latch1h<<8 | s->latch1l);
  else if RISING(timclk)
     s->time1 = (s->time1 - 1) & 0xFFFF;

  BundleEvt(time1, s->time1, s->dt_gt);

  /* Reload logic - SR then D-type */

  s->underflow1 = (s->time1==0);

  if HI(loadtx1)
     s->fload1 = TRUE;
  if HI(load1)
     s->fload1 = FALSE;

  if FALLING(timclk)
     SignalEvt(load1, (s->fload1 || s->underflow1), s->dt_lch);

  /* Write Latches */

  if HI(txll1)
     s->latch1l = BINVAL(data);

  if HI(txlh1)
     s->latch1h = BINVAL(data);

  /* Count latches */

  if HI(latchtx1) {
     s->count1h = (s->time1>>8)&0xFF;
     s->count1l = (s->time1)&0xFF;
     }

  /* Count reads */

  if HI(txch1)
    BundleEvt(data, s->count1h, s->dt_oe);
  if HI(txcl1)
    BundleEvt(data, s->count1l, s->dt_oe);

  /* Interupt output */

  if FALLING(timclk)
     SignalEvt(uftx1, s->underflow1, s->dt_gt);

/*--------------------------Interupt Registers-------------------------------*/

/*-----------------------------IRQ A-----------------------------------------*/

/* Write and Read the mask */

  if HI(reset) {
     s->imaska = 0;
     s->irqa  = 0x80; /* Top bit always set */
     s->irqal = 0x80;
/* Don't reset bit 4 of irqa, as this is the por flag */
     s->irqau = (s->irqau & 0x10) | 0x80;
     }

  if (HI(prog) && (s->dec_irqam) && HI(lwrite))
     s->imaska = BINVAL(data);

  if HI(dec_irqamr)
     BundleEvt(data, (s->imaska & 0xFD) , s->dt_lch);
  if FALLING(dec_irqamr)
     BundleZZZ(data, s->dt_oe);

/* Write the clear register (Note, no clear register in IRQ_B */

  if (HI(prog) && HI(lwrite) && s->dec_irqar) {
     s->irqa  = (s->irqa  & ~(0x7F & BINVAL(data))) | 0x80; /* Top bit unimplemented */
     s->irqau = (s->irqau & ~(0x7F & BINVAL(data))) | 0x80;
     s->irqal = (s->irqal & ~(0x7F & BINVAL(data))) | 0x80;
     }


/* Status Register Reads */

   if HI(dec_irqasr)
      BundleEvt(data, s->irqa, s->dt_lch);
   if FALLING(dec_irqasr)
      BundleZZZ(data, s->dt_oe);

/* Status register written from many sources - all edge sensitive and asynchronous */

   /* Bit 7 set high at reset */

   if RISING(uftx1)
      s->irqau = s->irqau | 0x40; /* Timer 1 - Bit 6 */ 

   if RISING(uftx0)
      s->irqau = s->irqau | 0x20; /* Timer 0 - Bit 5 */ 

   if HI(por)
      s->irqau = s->irqau | 0x10; /* Power on reset - Bit 4 */

   if RISING(flybi) {
      s->irqal = s->irqal | 0x08; /* Fly Back - Bit 3. Note: only one stage of sync needed */
      s->irqau = s->irqau | 0x08; /*  This bodge so that flybk does not get cleared in the latch */
      }

   if FALLING(Nindex)
      s->irqau = s->irqau | 0x04; /* Nindex - Bit 2 */

   /* Bit 1 does not exist */

   if RISING(pintr)
      s->irqau = s->irqau | 0x01; /* Printer Interrupt - Bit 0 */

/* Now synchronise these to rclk */

   if RISING(rclk)
      s->irqal = s->irqau;

   if LO(rclk)
      s->irqa = s->irqal;

/* Request Register Reads */

   if HI(dec_irqarr)
      BundleEvt(data, (s->irqa & s->imaska), s->dt_lch);
   if FALLING(dec_irqarr)
      BundleZZZ(data, s->dt_oe);

/* Merge all requests to make irqa */

   if (((s->irqa&0xFF) & s->imaska) != 0)
      SignalEvt(irqa, TRUE, s->dt_gt);
     else
      SignalEvt(irqa, FALSE, s->dt_gt);

/*-----------------------------IRQ B-----------------------------------------*/

/* Write and Read the mask */

  if HI(reset)
     s->imaskb = 0;
 
  if (HI(prog) && (s->dec_irqbm) && HI(lwrite))
     s->imaskb = BINVAL(data);
 
  if HI(dec_irqbmr)
     BundleEvt(data, s->imaskb, s->dt_lch);
  if FALLING(dec_irqbmr)
     BundleZZZ(data, s->dt_oe);

/* Status Register Reads */

  if HI(dec_irqbsr)
     BundleEvt(data, s->irqb, s->dt_gt);
  if FALLING(dec_irqbsr)
     BundleZZZ(data, s->dt_oe);

/* Synchronise all the level sensitive inputs */

  if RISING(rclk)
     s->irqbl = ((HI(kbrx)  <<7) |
                            (HI(kbtx)  <<6) |
                            (LO(Npirq) <<5) |
                            (LO(Nfintr)<<4) |
                            (LO(Niext) <<3) |
                            (LO(Nsintr)<<2) |
                            (LO(Nscirq)<<1) |
                             LO(Npfiq));

  if LO(rclk)
     s->irqb = s->irqbl;

/* Request Register Reads */

   if HI(dec_irqbrr)
      BundleEvt(data, (s->irqb & s->imaskb), s->dt_gt);
   if FALLING(dec_irqbrr)
      BundleZZZ(data, s->dt_oe);
 
/* Merge all requests to make irqb */

   if (((s->irqb&0xFF) & s->imaskb) != 0)
      SignalEvt(irqb, TRUE, s->dt_gt);
     else
      SignalEvt(irqb, FALSE, s->dt_gt);

/*---------------The Final IRQ Output----------------------------------------*/

  SignalEvt(Nirq, !(HI(irqa) || HI(irqb) || HI(dagint)), s->dt_gt);

/*------------------------------FIQ------------------------------------------*/

  if HI(reset)
     s->fiql = 0x80;

/* Write and Read the mask */
 
  if HI(reset)
     s->fmask = 0;

  if (HI(prog) && (s->dec_fiqm) && HI(lwrite))
     s->fmask = BINVAL(data);
 
  if HI(dec_fiqmr)
     BundleEvt(data, (s->fmask & 0xD3) , s->dt_lch); /* bits 5,3,2 = 0 */
  if FALLING(dec_fiqmr)
     BundleZZZ(data, s->dt_oe);
 
/* Status Register Reads */
 
  if HI(dec_fiqsr)
     BundleEvt(data, s->fiq, s->dt_gt);
  if FALLING(dec_fiqsr)
     BundleZZZ(data, s->dt_oe);


  /* Synchronise external FIQ sources */

  if RISING(rclk)
     s->fiql =  ( 1<<7 |  (LO(Npfiq) <<6) |
                          (LO(Nsintr)<<4) |
                          (LO(Niext) <<1) |
                           HI(fdrq));

  if LO(rclk)
     s->fiq = (s->fiql & 0xFF);;

/* Request Register Reads */
 
   if HI(dec_fiqrr)
      BundleEvt(data, (s->fiq & s->fmask), s->dt_gt);
   if FALLING(dec_fiqrr)
      BundleZZZ(data, s->dt_oe);
 

/* Merge all requests to make Nfiq */
 
   if (((s->fiq&0xFF) & s->fmask) != 0)
      SignalEvt(Nfiq, FALSE, s->dt_gt);
     else
      SignalEvt(Nfiq, TRUE, s->dt_gt);

/*-------------- All done --------------------------*/
}

static void IOINT_init(modelInst)
VoidStar modelInst;
{
  IOINT_state *s;
  IOINT_model me = (IOINT_model) modelInst;
  NEW(s, IOINT_state);
  me->model.proc = IOINT_proc;
  s->dt_gt = DELAY("gt");
  s->dt_oe = DELAY("oe");
  s->dt_lch = DELAY("lch");
  s->time0 = 0xFFFF;
  s->time1 = 0xFFFF;

  me->model.state = (VoidStar ) s;

}

void IOINTC_declare()
{
    simmodel_begin("iomd.iointc");
    simmodel_declare_s(IOINT_def, IOINT_init);
    simmodel_end();
}

/*--------------------------------------------------------------------------*/
