/* ############################################################ */
/* ###   KBDC - Keyboard C block                            ### */
/* ###                                                      ### */
/* ###   Change history :                                   ### */
/* ###   28/10/92 : WHO : Started                           ### */
/* ###   07/12/92 : 1.1 : made model timer like netlist so  ### */
/* ###                    vectors match                     ### */
/* ############################################################ */


#include <Handy.h>
#include "kbd.h"

#define  RESET  0x00
#define  WAIT_H 0x20
#define  IDLE   0x28
#define  TX1_64 0x12
#define  TX1    0x02
#define  RX1    0x04
#define  TX2_64 0x13
#define  TX2    0x03
#define  LOCK   0x39

/*

         ||                   inputs                      ||  next
  state  || res | clk | c12 | c11 | t64 | t2m | rxi | txi || state
---------++-----+-----+-----+-----+-----+-----+-----+-----++---------
  ANY    ||  1  |     |     |     |     |     |     |     || RESET

  RESET  ||  0  |     |     |     |     |     |     |     || WAIT_H

  WAIT_H ||  0  |  0  |     |     |     |     |     |     || WAIT_H
  WAIT_H ||  0  |  1  |     |     |     |     |     |     || IDLE

  IDLE   ||  0  |  1  |     |     |     |     |     |  1  || IDLE
  IDLE   ||  0  |  0  |     |     |     |     |     |  1  || RX1
  IDLE   ||  0  |     |     |     |     |     |     |  0  || TX1_64

  TX1_64 ||  0  |     |     |     |  0  |     |     |     || TX1_64
  TX1_64 ||  0  |     |     |     |  1  |     |     |     || TX1

  TX1    ||  0  |     |  0  |     |     |  0  |     |     || TX1
  TX1    ||  0  |     |  1  |     |     |  0  |     |     || IDLE
  TX1    ||  0  |     |     |     |     |  1  |     |     || RESET

  RX1    ||  0  |     |     |  0  |     |  0  |     |     || RX1
  RX1    ||  0  |     |     |  1  |     |  0  |     |     || LOCK
  RX1    ||  0  |     |     |     |     |  1  |     |     || RESET

  TX2_64 ||  0  |     |     |     |  0  |     |     |     || TX2_64
  TX2_64 ||  0  |     |     |     |  1  |     |     |     || TX2

  TX2    ||  0  |     |  0  |     |     |  0  |     |     || TX2
  TX2    ||  0  |     |  1  |     |     |  0  |  0  |     || WAIT_H
  TX2    ||  0  |     |  1  |     |     |  0  |  1  |     || LOCK
  TX2    ||  0  |     |     |     |     |  1  |     |     || RESET

  LOCK   ||  0  |     |     |     |     |     |  1  |  1  || LOCK
  LOCK   ||  0  |     |     |     |     |     |     |  0  || TX2_64
  LOCK   ||  0  |     |     |     |     |     |  0  |  1  || WAIT_H

         || tim | clk | cnt |     |     |     ||
  state  || res |  oe | res | rxb | txb | srx ||
---------++-----+-----+-----+-----+-----+-----++
  RESET  ||  0  |  0  |  0  |  0  |  0  |  0  || 00
  WAIT_H ||  1  |  0  |  0  |  0  |  0  |  0  || 20
  IDLE   ||  1  |  0  |  1  |  0  |  0  |  0  || 28
  TX1_64 ||  0  |  1  |  0  |  0  |  1  |  0  || 12
  TX1    ||  0  |  0  |  0  |  0  |  1  |  0  || 02
  RX1    ||  0  |  0  |  0  |  1  |  0  |  0  || 04
  TX2_64 ||  0  |  1  |  0  |  0  |  1  |  1  || 13
  TX2    ||  0  |  0  |  0  |  0  |  1  |  1  || 03
  LOCK   ||  1  |  1  |  1  |  0  |  0  |  1  || 39

*/


#define KBDC_def "KBDC(enable,test1,test2,reset,clk,clkin,krxint,ktxint,cclkoe,cdataoe,dataoe) => \
(t16,Nkclkoe,Nkdataoe,rxb,txb,srx,stx,st[7:0],ccnt[3:0],nostx)"

typedef struct {
  ModelCore model;

  Signal *enable,*test1,*test2,*reset,*clk,*clkin,*krxint,*ktxint,*cclkoe,*cdataoe,*dataoe;
  Signal *t16,*Nkclkoe,*Nkdataoe,*rxb,*txb,*srx,*stx;
  Bundle *st,*ccnt;
  Signal *nostx;
} ModelOfKBDC;

typedef struct {
  int state,pre_time,mid_time,end_time,clk_count;
} KBDCState;

static void KBDC_proc(ModelInst)
VoidStar ModelInst;
{

  ModelOfKBDC *me = (ModelOfKBDC *)ModelInst;
  KBDCState *s = me->model.state;
  int res_tim,res_cnt,c11,c12,us_64,ms_16,newbit;

  if (s->state&0x20) res_tim=1;
  else               res_tim=0;

  if (s->state&0x08) res_cnt=1;
  else               res_cnt=0;

  if RISING(clk) {                                  /* the timer clock */
    if      (res_tim)           s->pre_time=0;
    else if (s->pre_time==0xFF) s->pre_time=0;
    else                        s->pre_time++;
  }
  if RISING(clk) {
    if       (res_tim)                         s->mid_time=0;
    else if  (s->mid_time==0x03)               s->mid_time=0;
    else if ((s->pre_time==0xFF) || HI(test1)) s->mid_time++;
  }
  if RISING(clk) {
    if      (res_tim)                                        s->end_time=0;
    else if (s->end_time==0xFF)                              s->end_time=0;
    else if (((s->pre_time==0xFF) && (s->mid_time==0x03)) ||
             ((s->pre_time==0xFF) && HI(test2))           ||
             (HI(test1) && HI(test2)))                       s->end_time++;
  }

  us_64 = (s->mid_time & 0x02) ? 1 : 0;
  ms_16 = (s->end_time & 0x80) ? 1 : 0;

  if LO(test1) {
     if (s->pre_time & 0x80) SignalEvt(t16, ONE,300);
     else                    SignalEvt(t16,ZERO,300);
  }
  else                       SignalEvt(t16,VAL(clk),300);

  if      (res_cnt)     s->clk_count=0;    /* the Keyboard clock counter for C11 and C12 */
  else if (RISING(clkin) && (me->clkin->value==ONE)) s->clk_count++;    /* asynchronous reset                         */

  if (s->clk_count==11) c11=1;             /* decode a count of 11                       */
  else                  c11=0;

  if (s->clk_count==12) c12=1;             /* decode a count of 12                       */
  else                  c12=0;

  if LO(enable) {
     SignalEvt(nostx,ZERO,1);
  }
  else if RISING(clk) {
    if ((s->state==TX2) && !ms_16 && c12 && HI(krxint)) SignalEvt(nostx,ONE,1);
    else if ((s->state!=LOCK) && (s->state!=WAIT_H))    SignalEvt(nostx,ZERO,1);
  }

  if HI(reset) {
    s->state=RESET;
    SignalEvt(stx,ZERO,1);
  }
  else if RISING(clk) {
    if (LO(enable)) {
           s->state=RESET;   /* synchronous reset */
           SignalEvt(stx,ZERO,1);
    }
    else {
       if (LO(nostx) && (((s->state==WAIT_H) && HI(clkin)) ||
           ((s->state==TX1) && !ms_16 && c12)  ||
           ((s->state==TX2) && !ms_16 && c12 && HI(krxint)))) SignalEvt(stx, ONE,1);
       else                                                   SignalEvt(stx,ZERO,1);
       switch(s->state) {
         case RESET:  if      (LO(reset) && HI(enable)      ) s->state=WAIT_H;
                      break ;
         case WAIT_H: if       HI(clkin)                      s->state=IDLE;
                      break;
         case IDLE:   if      (LO(clkin) && HI(ktxint)      ) s->state=RX1;
                      else if (             LO(ktxint)      ) s->state=TX1_64;
                      break;
         case TX1_64: if       (us_64)                        s->state=TX1;
                      break;
         case TX1:    if      ((!ms_16) && c12              ) s->state=IDLE;
                      else if (( ms_16)                     ) s->state=WAIT_H;
                      break;
         case RX1:    if      ((!ms_16) && c11              ) s->state=LOCK;
                      else if (( ms_16)                     ) s->state=WAIT_H;
                      break;
         case TX2_64: if       (us_64)                        s->state=TX2;
                      break;
         case TX2:    if      ((!ms_16) && c12 && HI(krxint)) s->state=LOCK;
                      else if ((!ms_16) && c12 && LO(krxint)) s->state=WAIT_H;
                      else if (( ms_16)                     ) s->state=WAIT_H;
                      break;
         case LOCK:   if      (              LO(ktxint)     ) s->state=TX2_64;
                      else if (LO(krxint) && HI(ktxint)     ) s->state=WAIT_H;
                      break;
         default:     errhandler_f1(SS_MODELS+DISASTER,"%s: Unknown state\n",(int)INSTNAME);
       }
    }
  }

  if ((s->state&0x10) || (HI(cclkoe))) SignalEvt(Nkclkoe,ZERO, 300);
  else                                 SignalEvt(Nkclkoe, ONE, 300);

/*  if (((s->state&0x08) && HI(enable)) ||  */
  if (((s->state&0x02) && HI(dataoe)) ||
      (HI(cdataoe)))                    SignalEvt(Nkdataoe,ZERO, 300);
  else                                  SignalEvt(Nkdataoe, ONE, 300);

  if (s->state&0x04) SignalEvt(rxb, ONE, 300);
  else               SignalEvt(rxb,ZERO, 300);

  if (s->state&0x02) SignalEvt(txb, ONE, 300);
  else               SignalEvt(txb,ZERO, 300);

  if (s->state&0x01) SignalEvt(srx, ONE,300);
  else               SignalEvt(srx,ZERO,300);

  BundleEvt(st,s->state,1);
  BundleEvt(ccnt,s->clk_count,1);
}

static void KBDC_init(ModelInst)
VoidStar ModelInst;
{
  KBDCState *s;
  ModelOfKBDC *me = (ModelOfKBDC *)ModelInst;
  NEW(s, KBDCState);
  me->model.proc = KBDC_proc;
  s->state=RESET;
  s->clk_count=0;
  SignalEvt(stx,ZERO,1);
  BundleEvt(st,s->state,1);
  me->model.state = (VoidStar )s;
}


void KBDC_declare()
{
  simmodel_begin("iomd.kbdc");
  simmodel_declare_s(KBDC_def, KBDC_init);
  simmodel_end();
}
