/* vctlc.c */
/* ############################################################### */
/* ############################################################### */
/* ###                                                         ### */
/* ###  VCTL block level model                                 ### */
/* ###                                                         ### */
/* ###  Created 27/8/92: Simon Segars                          ### */
/* ###                                                         ### */
/* ###                                                         ### */
/* ############################################################### */
/* ############################################################### */
/* ###  Revision History:                                      ### */
/* ###   27/8/92: SAS : Initial release                        ### */
/* ###    1/9/92: SAS : FAST and SLOW state machines           ### */
/* ###   12/1/93: SAS : Post review mods                       ### */
/* ###   19/01/93 WHO : changed sc_en clock from vclk to pclk  ### */
/* ###   28/01/93 WHO : made reset state of ssm = 2, so Ncdoe=0    */
/* ############################################################### */


#include <Handy.h>
#include "vctl.h"
#include "busarb.h"

/* ############################################################### */

typedef struct
{
  ModelCore     model;
  Signal        *dummy;

} *DelayModel;

static int get_delay (delay_name, model)
char *delay_name;
VoidStar model;
{
  Parameter    param;
  int          delay = 1000*NS;
  DelayModel   me = (DelayModel)model;

  param = simmodel_param(me->model.state,"delay");
  if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  if (simmodel_type(param) == PT_PARAMBLOCK) {
    param = simmodel_sub_param(param,delay_name);
    if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  }
  errhandler_f (SS_MODELS+DISASTER,"!!! %s delay not correctly specified\n",delay_name);
  return delay;
}

#define VCTL_def \
"VCTL(dmach[2:0],fsize[7:0],clk64,clk32,dmagoing,dmavbus,flybk,flybks,fsizewt,Ninit, Ninitu,Nsndrq,\
      Nvidrq,pclk,qsf,rclk,reset, sc, snden, vbusrq, vbusrqram, vbusrqprog, vidakn, viden, vnc,vncs,\
      vncd,vrm,was8ck,was9ck,wasFck) \
     => (cursrq, flybi, flybid, last, Ncdoe, Nse, Nsndak, Nvidak, snd0drq, split, transrq, vbusak, vclk, \
        tsm[1:0], ste[3:0], sts[1:0],ak,even,go,req)"

typedef struct {
    ModelCore	model;
    Bundle      *dmach, *fsize;
    Signal	*clk64, *clk32, *dmagoing, *dmavbus,*flybk, *flybks, *fsizewt, *Ninit, *Ninitu, *Nsndrq;
    Signal      *Nvidrq, *pclk, *qsf, *rclk, *reset;
    Signal      *sc, *snden, *vbusrq, *vbusrqram, *vbusrqprog, *vidakn, *viden, *vnc, *vncs, *vncd, *vrm;
    Signal      *was8ck,*was9ck,*wasFck;
    Signal   	*cursrq, *flybi, *flybid, *last, *Ncdoe, *Nse, *Nsndak, *Nvidak, *snd0drq, *split;
    Signal      *transrq, *vbusak, *vclk;
    Bundle      *tsm, *ste, *sts;
    Signal      *ak,*even,*go, *req;
    } *VCTL_model;

typedef struct {
    int dt_gt;
    int fsm_state, ssm_state, lw0;
    int even, sc_en, vreq, vreql, vrequ, v_go;
    int half_rq, half_rq_u, tsm_state, full_rq, fcount;
    int vclkl, rclkl, pclk_en, vbusrql, dmaak;
    int vncl, flybkl;
    int vidaknl,was8,was9,wasF;
    int srq, crq;
    } VCTL_state;

/*------------------------------------------------------------------------*/

static void VCTL_proc (modelInst)
VoidStar modelInst;
{
    VCTL_model me = (VCTL_model) modelInst;
    VCTL_state *s = (VCTL_state *)me->model.state;

    int rvreq;

/* Merge the bus requests */

SignalEvt(vbusrq, ((HI(vbusrqram) && HI(dmagoing) && HI(dmavbus)) | HI(vbusrqprog)), s->dt_gt);

/*------------- 64Meg State Machine --------------------------------------*/
/* Note : this state machine has async. reset with Ninit for test only */

/* Synchronise Ninit to 64MHz */

  if RISING(clk64)
     SignalEvt(Ninit, HI(Ninitu), s->dt_gt);

  if LO(Ninit)
     s->fsm_state = 0;
  if HI(reset)
     s->fsm_state = (s->fsm_state & 0x1);

  if (FALLING(clk64) && HI(Ninit)) {
     switch(s->fsm_state) {
       case STATE_W0 : if (s->vreq && (s->ssm_state==STATE_Y) && LO(vbusrq))
                          s->fsm_state = STATE_A;
                         else
                          s->fsm_state = STATE_W1;
                       break;
       case STATE_W1 : s->fsm_state = STATE_W0;
                       break;
       case STATE_A  : if (s->v_go)
                          s->fsm_state=STATE_B;
                         else
                          s->fsm_state=STATE_F;
                       break;
       case STATE_B  : s->fsm_state = STATE_C;
                       break;
       case STATE_C  : s->fsm_state = STATE_D;
                       break;
       case STATE_D  : s->fsm_state = STATE_E;
                       break;
       case STATE_E  : s->fsm_state = STATE_F;
                       break;
       case STATE_F  : if (HI(vbusrq) && !(HI(last) || s->v_go))
                          s->fsm_state = STATE_W1; 
                       else 
                          s->fsm_state = STATE_A;
                       break;
       default : /* Should only get here if uninitialised */
                 s->fsm_state = STATE_W0; /* goto reset state*/
       }
     }

/* Latched version of STATE_W0 so that 32M s/m can move out of STATE_I */
  if HI(clk64)
     s->lw0 = (s->fsm_state==STATE_W0);

/* 2-stage sync of Nvidrq with clk32 */

  rvreq = (HI(vrm) && LO(Nvidrq) && HI(vnc));

/* Second Stage, D-type with reset */
   if (rvreq==FALSE) /* Reset */
      s->vreql = FALSE;
    else if FALLING(clk32)
     s->vreql = s->vrequ;

/* First stage, D-type without reset */
  if FALLING(clk32)
     s->vrequ = rvreq;
 
/* Finally all munged together */
  s->vreq = (s->vreql && rvreq);

  if HI(clk64) {
     s->was8 = (s->fsm_state==STATE_D);
     s->was9 = (s->fsm_state==STATE_A);
     s->wasF = (s->fsm_state==STATE_E);
     }

  SignalEvt(was8ck, s->was8 && LO(clk64), 2*s->dt_gt);
  SignalEvt(was9ck, s->was9 && LO(clk64), 2*s->dt_gt);
  SignalEvt(wasFck, s->wasF && LO(clk64), 2*s->dt_gt);

/* State machine outputs cleared on reset */

   if HI(reset) {
      SignalEvt(vidakn, TRUE, s->dt_gt);
      SignalEvt(Nsndak, TRUE, s->dt_gt);
      SignalEvt(Nvidak, TRUE, s->dt_gt);
      s->sc_en = 0;
      s->even = 0;
      }

/* D-type on even also */
   if (RISING(was9ck) && s->v_go)
      s->even = !s->even;

/* D-type with clear on v_go */
  if (HI(was9ck) && (s->vreq))
     s->v_go = 1;
  if (HI(reset) || HI(last))
     s->v_go = 0;

/* D-type on last */
   if RISING(wasFck)
      SignalEvt(last, (!s->vreq && !s->even), s->dt_gt);
   if (HI(reset) || HI(was9ck))
      SignalEvt(last, FALSE, s->dt_gt);

/* SRFF on Nvidak */
   if (HI(was9ck) && (s->vreq))
      SignalEvt(vidakn, FALSE, s->dt_gt);;
   if (HI(wasFck) && (!s->vreq && !s->even))
      SignalEvt(vidakn, TRUE, s->dt_gt);

   if FALLING(clk64)
      s->vidaknl = HI(vidakn);

  if (HI(vnc) && HI(vrm))
     SignalEvt(Nvidak, (HI(vidakn) || s->vidaknl), s->dt_gt);

/* sc_en is a pipelinned version of Nvidak */
   if FALLING(pclk)    /* was FALLING(vclk) WHO 19/1/93 */
      s->sc_en = (LO(Nvidak) && HI(vnc));

/*------------Primary State Machine Outputs --------------------*/

     SignalEvt(vclk, ((s->fsm_state&0x2)==0x2), s->dt_gt);
     SignalEvt(sc,     (s->sc_en && ((s->fsm_state&0x2)==0x2)),     s->dt_gt);
     SignalEvt(clk32,  s->fsm_state&0x1,          s->dt_gt);
     SignalEvt(vbusak, ((s->fsm_state&0x8)!=0x8), s->dt_gt);
     if FALLING(clk32)
        SignalEvt(Nse,    ((s->fsm_state&0x8)!=0x8), s->dt_gt);

/*------------------32Meg State Machine-------------------------*/

  if HI(reset)
     s->ssm_state = 2;
  else if RISING(clk32) {
    switch(s->ssm_state) {
      case STATE_I : if (s->lw0)
                        s->ssm_state = STATE_X;
                     break;
      case STATE_X : if (s->vreq && LO(vbusrq) && HI(rclk))
                        s->ssm_state = STATE_Y;
                     break;
      case STATE_Y : s->ssm_state = STATE_I;
                     break;
      default      : /* Should only get here when un-initialised */
                     s->ssm_state = STATE_X;  /* goto reset state */
      }
   }

/*------------ PCLK MUX and Buffer Control ---------------------*/

    SignalEvt(Ncdoe, (s->ssm_state==STATE_I), s->dt_gt);
    if FALLING(clk32)
       s->pclk_en = (s->ssm_state==STATE_X);

    if FALLING(clk64)
       s->rclkl = HI(rclk);

    if HI(clk64)
       s->vclkl = HI(vclk);

    if (s->pclk_en)
       SignalEvt(pclk, s->rclkl, s->dt_gt); /*pclk=rclk*/
      else
       SignalEvt(pclk, s->vclkl, s->dt_gt); /*pclk=vclk*/

/*------------ Synchronisers ------------------------------*/

   if FALLING(rclk) {  /* D-Types without reset */
      s->vncl = HI(vnc);
      s->flybkl = HI(flybk);
      SignalEvt(cursrq, (LO(flybk) && HI(viden) && !(HI(Nvidrq)||s->crq) && !(HI(vnc) && HI(vrm))),
                s->dt_gt); /*cursdrq masked out during flybk */
      SignalEvt(snd0drq, (HI(snden) && !(HI(Nsndrq)||s->srq) && !(HI(dmagoing) && 
               (BINVAL(dmach)==SND0DMA))), s->dt_gt);
      s->crq = HI(Nvidrq); /* 1st stage of vidrq sync */
      s->srq = HI(Nsndrq); /* 1st stage of sndrq sync */
      }


  if HI(rclk) {
     SignalEvt(vncs, s->vncl, s->dt_gt);
     SignalEvt(flybks, s->flybkl, s->dt_gt);
     }

/*----------------Transfer Request stuff---------------------*/

/*------ First the state machine (must get evaluated first) ------------*/

  if (LO(viden) || LO(vrm))
     s->tsm_state = 0;
   else if FALLING(rclk) {
     switch (s->tsm_state) {
        case STATE_IT : if (s->full_rq)
                           s->tsm_state = STATE_FF;
                        if (s->half_rq && LO(flybk))
                           s->tsm_state = STATE_SP;
                      break;
        case STATE_FF : if ((BINVAL(dmach)==TRANDMA) && HI(dmagoing))
                           s->tsm_state = STATE_S0;
                      break;
        case STATE_S0 : s->tsm_state = STATE_SP;
                      break;
        case STATE_SP : if ((BINVAL(dmach)==TRANDMA) && HI(dmagoing))
                           s->tsm_state = STATE_IT;
                      break;
        default       : s->tsm_state = STATE_IT; /* Should never get here */
        }
      }

/* Now the request generators */

  if HI(reset)
     s->fcount = (s->fcount | 0x80);

  if (HI(fsizewt) || (s->full_rq))
     s->fcount=BINVAL(fsize); 

  if (HI(flybi) && RISING(vncs)) 
     s->fcount--;

  if CHANGING(qsf) /* Twin opposite edged D-types with reset */
     s->half_rq_u = TRUE;
  if (HI(reset) || (HI(dmagoing) && HI(rclk) && (BINVAL(dmach)==TRANDMA)))
     s->half_rq_u = FALSE;

  if FALLING(rclk) {
     s->full_rq = (s->fcount==0);  /* Full request, sync'd to rclk */
     if (s->fcount==0)
        SignalEvt(flybi, FALSE, s->dt_gt);
     s->half_rq = s->half_rq_u;    /* Split request, sync'd to rclk */
     }

  if RISING(flybks)
     SignalEvt(flybi, TRUE, s->dt_gt);
  if LO(flybk)
     SignalEvt(flybi, FALSE, s->dt_gt);

  if HI(rclk) 
     SignalEvt(split, (s->tsm_state==STATE_SP), s->dt_gt);

  SignalEvt(transrq, ((s->tsm_state&0x1)==0x1), s->dt_gt);
  BundleEvt(tsm, s->tsm_state, s->dt_gt);

/* -------Cursor and Sound DMA Acknowledge-------------------- */

  if FALLING(rclk)
    s->vbusrql = HI(vbusrq);

  s->dmaak = !(HI(vbusak) && s->vbusrql && (s->ssm_state==STATE_X));

  if (BINVAL(dmach)==CURSDMA)
     SignalEvt(Nvidak, s->dmaak, s->dt_gt);
  if (BINVAL(dmach)==SND0DMA)
     SignalEvt(Nsndak, s->dmaak, s->dt_gt);

  if RISING(clk32) {
     SignalEvt(flybid,VAL(flybi),s->dt_gt);
     SignalEvt(vncd,  VAL(vncs), s->dt_gt);
  }

/*----------- All Done --------------------------------*/
BundleEvt(ste, s->fsm_state, s->dt_gt);
BundleEvt(sts, s->ssm_state, s->dt_gt);
SignalEvt(go, s->v_go, s->dt_gt);
SignalEvt(req, s->vreq, s->dt_gt);
SignalEvt(even, s->even, s->dt_gt);
}

/*------------------------------------------------------------------------*/

static void VCTL_init(modelInst)
VoidStar modelInst;
{
  VCTL_state *s;
  VCTL_model me = (VCTL_model) modelInst;
  NEW(s, VCTL_state);
  me->model.proc = VCTL_proc;
  s->dt_gt = DELAY("gt");

/* Reset these to random values to check robustness */
  me->model.state = (VoidStar ) s;

}

void VCTLC_declare()
{
    simmodel_begin("iomd.vctlc");
    simmodel_declare_s(VCTL_def, VCTL_init);
    simmodel_end();
}

